<?php

namespace Drupal\calendar_view\Plugin\views\style;

use Drupal\Component\Utility\Xss;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItem;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Drupal\datetime_range\Plugin\Field\FieldType\DateRangeItem;
use Drupal\views\Plugin\views\filter\Date as DateFilter;
use Drupal\views\Plugin\views\style\DefaultStyle;
use Drupal\views\ResultRow;

/**
 * Unformatted style plugin to render rows one after another with no
 * decorations.
 *
 * @ingroup views_style_plugins
 *
 * @ViewsStyle(
 *   id = "calendar",
 *   title = @Translation("Calendar"),
 *   help = @Translation("Displays rows in a calendar."),
 *   theme = "views_view_calendar",
 *   display_types = {"normal"}
 * )
 */
class Calendar extends DefaultStyle {

  /**
   * Retrieve all Date fields exposed as filters.
   *
   * @return array
   *   List of exposed filter names, keyed by field ID.
   */
  public function getDateFilters() {
    // Improve performance with static variables.
    $date_filters = &drupal_static(__FUNCTION__);
    if (isset($date_filters)) {
      return $date_filters;
    }

    $date_filters = [];
    $filters = $this->view->display_handler->getHandlers('filter');
    foreach ($filters as $filter_id => $filter) {
      $date_filters[$filter_id] = $filter instanceof DateFilter ? $filter : NULL;
    }
    return array_filter($date_filters);
  }

  /**
   *
   */
  public function getCalendarTimestamp() {
    // Avoid unnecessary calls with static variable.
    $default_time = &drupal_static(__FUNCTION__);
    if (isset($default_time)) {
      return $default_time;
    }

    // Get date (default: today).
    $timestamp = $this->view->getExposedInput()['calendar_timestamp'] ?? NULL;
    $timestamp = empty($timestamp) ? date('U') : $timestamp;

    // Allow user to pass query string (e.g "<url>?calendar_timestamp=2022-12-31").
    if (!ctype_digit($timestamp)) {
      $timestamp = strtotime($timestamp);
    }

    return $timestamp;
  }

  /**
   * Get default options, statically.
   *
   * @return array
   */
  public static function getDefaultOptions() {
    return [
      'calendar_filters' => [],
      'calendar_weekday_start' => 1, // Monday
      'calendar_sort_order' => 'ASC',
      'calendar_timestamp' => 'this month',
    ];
  }

  /**
   * {@inheritdoc}
   */
  protected function defineOptions() {
    $options = parent::defineOptions();

    $defaults = self::getDefaultOptions();
    foreach ($defaults as $key => $value) {
      $options[$key] = ['default' => $value];
    }

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state) {
    parent::buildOptionsForm($form, $form_state);

    $date_filters = $this->getDateFilters();
    $date_filter_keys = array_keys($date_filters);
    $default_date_filter = [reset($date_filter_keys)];

    $form['calendar_filters'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Date fields'),
      '#empty_option' => $this->t('- Select -'),
      '#options' => array_combine($date_filter_keys, $date_filter_keys),
      '#default_value' => $this->options['calendar_filters'] ?? $default_date_filters,
      '#disabled' => empty($date_filters),
    ];
    if (empty($date_filters)) {
      $form['calendar_filters']['#description'] = $this->t('Empty? You need to add a date field in exposed filters on this View.');
    }

    $form['calendar_weekday_start'] = [
      '#type' => 'select',
      '#title' => $this->t('Start week on:'),
      '#options' => [
        1 => t('Monday'),
        2 => t('Tuesday'),
        3 => t('Wednesday'),
        4 => t('Thursday'),
        5 => t('Friday'),
        6 => t('Saturday'),
        0 => t('Sunday'),
      ],
      '#default_value' => $this->options['calendar_weekday_start'] ?? 1,
    ];

    $form['calendar_sort_order'] = [
      '#type' => 'select',
      '#title' => $this->t('Default sort order'),
      '#options' => [
        'ASC' => $this->t('Chronological'),
        'DESC' => $this->t('Antichronological'),
      ],
      '#default_value' => $this->options['calendar_sort_order'] ?? 'ASC',
      '#required' => TRUE,
    ];

    $form['calendar_timestamp'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Default date'),
      '#description' => $this->t('Use any date format but, if possible, a timestamp is better.'),
      '#default_value' => $this->options['calendar_timestamp'] ?? 'this month',
    ];
  }

  /**
   * Where the magic happens.
   *
   * @param string $filter_id
   *   The name of an exposed date filter on this View.
   * @param array $result
   *   An array of ResultRow values.
   */
  public function processResults(string $filter_id, array $results = []) {
    $filter = $this->getDateFilters()[$filter_id] ?? NULL;
    if (!$filter) {
      return [];
    }

    $content = [];
    $min = $max = NULL;

    $inputs = $this->view->getExposedInput();

    // Respect filters configuration and user input.
    // Use an array "between" by default to support advanced filtering.
    $operator = $filter->operator ?? 'between';
    $default_value = $filter->value['value'] ?? [];
    $input = $inputs[$filter_id] ?? $default_value;
    if (!empty($input)) {
      if (\is_array($input)) {
        $input_min = $input['min'] ?? NULL;
        $input_max = $input['max'] ?? NULL;
        $input_start = $input_end = NULL;
        if (!empty($input_min)) {
          $input_time_min = !ctype_digit($input_min) ? strtotime($input_min) : Xss::filter($input_min);
          $input_start = strtotime(date('Y-m-d 00:00:00', $input_time_min));
        }
        if (!empty($input_max)) {
          $input_time_max = !ctype_digit($input_max) ? strtotime($input_max) : Xss::filter($input_max);
          $input_end = strtotime(date('Y-m-d 23:59:59', $input_time_max));
        }
      } else {
        $input_time = !ctype_digit($input) ? strtotime($input) : Xss::filter($input);
        $input_start = strtotime(date('Y-m-d 00:00:00', $input_time));
        $input_end = strtotime(date('Y-m-d 23:59:59', $input_time));
      }

      switch ($operator) {
        case 'between':
        case '<>':
        case '!=':
        case '=':
          $input = ['min' => $input_start, 'max' => $input_end];
          break;
        case '>=':
          $input = ['min' => $input_start, 'max' => NULL];
          break;
        case '>':
          $input = ['min' => $input_start + 1, 'max' => NULL];
          break;
        case '<=':
          $input = ['min' => NULL, 'max' => $input_end];
          break;
        case '<':
          $input = ['min' => NULL, 'max' => $input_end - 1];
          break;
        default:
          $input = ['min' => NULL, 'max' => NULL];
          break;
      }
    } else {
      $input = ['min' => NULL, 'max' => NULL];
    }
    foreach ($input as $key => $value) {
      if (!empty($value) && !ctype_digit($value)) {
        $input[$key] = strtotime($value);
      }
    }
    $min = $input['min'] ?? NULL;
    $max = $input['max'] ?? NULL;

    foreach ($results as &$result) {
      // @todo How to deal with _relationship_entities?
      $entity = $result instanceof ResultRow ? $result->_entity : NULL;

      // Get correct entity field's name.
      $field_name = $filter->configuration['field_name'] ??
      $filter->configuration['entity field'] ??
      str_replace('_value', '', $filter_id);

      try {
        if (!$entity instanceof EntityInterface) {
          throw new \Exception('Entity not found.');
        }
        if (!$entity->hasField($field_name)) {
          throw new \Exception('Field does not exist: ' . $field_name);
        }
        if (($items = $entity->get($field_name))->isEmpty()) {
          throw new \Exception('Field is empty: ' . $field_name);
        }
        if (!$items) {
          throw new \Exception('Field items undefined.');
          continue;
        }

        foreach (range(1, $items->count()) as $i) {
          $item = $items->get($i - 1);

          $values = [];

          // Default date field values.
          $timestamp = $item->value;
          $values['from'] = $values['to'] = $timestamp;
          $values['item'] = $item;
          $values['calendar_filter'] = $field_name;

          // Datetime field.
          if ($item instanceof DateTimeItemInterface) {
            $datetime_type = $item->getFieldDefinition()->getSetting('datetime_type');
            $values['datetime'] = $datetime_type;

            // Get date(time) or range.
            if (is_a($item, DateRangeItem::class, TRUE)) {
              $values['from'] = $item->get('value')->getDateTime()->getTimestamp();
              $values['to'] = $item->get('end_value')->getDateTime()->getTimestamp();
            } elseif ($item instanceof DateTimeItem) {
              $values['from'] = $values['to'] = $item->date->getTimestamp();
            }
          }

          // Insert content occurences.
          if ($from = $values['from'] ?? NULL) {
            $time_date = date('Y-m-d 00:00:00', $from);
            $timestamp = (int) date('U', strtotime($time_date));
            // Respect limitations.
            $is_min_ok = (!$min || $min && $timestamp >= $min);
            $is_max_ok = (!$max || $max && $timestamp <= $max);
            if ($is_min_ok && $is_max_ok) {
              $content[$timestamp][] = $result;
            }
          }
          if ($to = $values['to'] ?? NULL) {
            $date_time_from = new \DateTime();
            $date_time_from->setTimestamp($from);

            $date_time_to = new \DateTime();
            $date_time_to->setTimestamp($to);

            $interval = $date_time_to->diff($date_time_from);

            // Calculcate days span.
            $start = $interval->d == 0 ? 1 : 0;
            for ($i = $start; $interval->d >= $i; $i++) {
              $date_time_from->modify('+1 day');
              $time_date = date('Y-m-d 00:00:00', $date_time_from->format('U'));
              $timestamp = (int) date('U', strtotime($time_date));

              // Respect limitations.
              $is_min_ok = (!$min || $min && $timestamp >= $min);
              $is_max_ok = (!$max || $max && $timestamp <= $max);
              if ($is_min_ok && $is_max_ok) {
                $content[$timestamp][] = $result;
              }
            }
          }

          // Keep track of thing for later use.
          // @see template_preprocess_calendar_view_day()
          $result->calendar_view = $values;
        }
      } catch (\Exception$e) {
        // Fail silently.
        $content['errors'][] = $result;
      }
    }

    // Sort content by timestamp.
    $sort_order = $this->options['calendar_sort_order'] ?? 'ASC';
    if ($sort_order == 'DESC') {
      krsort($content);
    } else {
      ksort($content);
    }

    return $content;
  }

  /**
   * {@inheritDoc}
   */
  public function preRender($result) {
    parent::preRender($result);

    // Calendars might already been built.
    // The "navigation by month" pager for instance is called first by ViewExecutable.
    // See \Drupal\calendar_view\Plugin\views\pager\Calendar::preRender().
    if (isset($this->view->calendars)) {
      return;
    }

    // Prepare calendar.
    $this->view->calendars = $this->buildCalendars($result);
  }

  /**
   * {@inheritDoc}
   */
  public function render() {
    // Add default cache tags to Calendars.
    $cache_tags = $this->view->getCacheTags() ?? [];
    foreach ($this->view->calendars ?? [] as &$calendar) {
      $calendar['#cache']['contexts'] = ['url.query_args:calendar_timestamp'];
      $calendar['#cache']['tags'] = $cache_tags;
    }

    return parent::render();
  }

  /**
   * Render a calendar.
   *
   * @todo How to start on another day?
   * @todo Make weekday label configurable.
   * @todo How to trim filter field name correctly?
   */
  public function buildCalendars($result) {
    // Check which exposed Date filters were selected.
    $available_date_filters = $this->getDateFilters();
    $calendar_filters = $this->options['calendar_filters'] ?? [];
    $calendar_filters = array_filter($calendar_filters, function ($field_name) use ($available_date_filters) {
      return ($field_name !== 0) && isset($available_date_filters[$field_name]);
    });
    if (!$calendar_filters) {
      if (!empty($available_date_filters)) {
        $keys = array_keys($available_date_filters);
        $default_calendar_filter_id = reset($keys);
        $this->options['calendar_filters'] = $calendar_filters = [$default_calendar_filter_id];
      } else {
        // Zero date fields available.
        return;
      }
    }

    $contents = [];
    foreach ($calendar_filters as $filter_id) {
      // Process results for calendar display.
      $list = $this->processResults($filter_id, $result);
      foreach ($list as $timestamp => $rows) {
        if (!isset($contents[$timestamp])) {
          $contents[$timestamp] = [];
        }
        $contents[$timestamp] += $rows;
      }
    }

    // Make sure timestamps are valid.
    $timestamps = array_filter(array_keys($contents), function ($value) {
      return ctype_digit($value);
    });

    // Use contextual timestamp, set in URl query or arguments.
    // @see \Drupal\calendar_views\Plugin\views\pager\Calendar::preRender()
    if ($selected_timestamp = $this->view->calendar_timestamp ?? NULL) {
      $selected_year = date('Y', $selected_timestamp);
      $selected_month = date('m', $selected_timestamp);
      $selected_month_start = strtotime($selected_year . '-' . $selected_month . '-' . '01');
      $selected_month_days = date('t', $selected_month_start);
      $selected_month_end = strtotime($selected_year . '-' . $selected_month . '-' . $selected_month_days);

      $timestamps = array_filter(array_keys($contents), function ($value) use ($selected_month_start, $selected_month_end) {
        return $value >= $selected_month_start && $value <= $selected_month_end;
      });
    }

    // Render an empty calendar if no results.
    if (empty($timestamps)) {
      $default_time = $selected_timestamp ?? $this->options['calendar_timestamp'] ?? date('U');
      if (!ctype_digit($default_time)) {
        $default_time = strtotime($default_time);
      }
      $timestamps = [$default_time];
    }

    // Build calendars.
    $calendars = [];
    foreach ($timestamps as $timestamp) {
      $year = date('Y', $timestamp);
      $month = date('m', $timestamp);
      if (!isset($calendars[$year . $month])) {
        $calendars[$year . $month] = $this->buildMonthTable($year, $month);
      }
    }

    // Populate calendars.
    foreach ($calendars as $key => $table) {
      foreach ($table['#rows'] ?? [] as $i => $rows) {
        foreach ($rows['data'] ?? [] as $timestamp => $cells) {
          $results = [];
          foreach ($contents[$timestamp] ?? [] as $index => $result) {
            $renderable_row = $this->view->rowPlugin->render($result);
            $results[] = $renderable_row;
          }

          $line = &$calendars[$key]['#rows'][$i];
          $cell = &$line['data'][$timestamp];
          $cell['data']['#children'] = $results;
        }
      }
    }

    return $calendars;
  }

  /**
   * Render a month calendar as a table.
   */
  public function buildMonthTable($year, $month) {
    // Build table.
    $days = [
      1 => t('Monday'),
      2 => t('Tuesday'),
      3 => t('Wednesday'),
      4 => t('Thursday'),
      5 => t('Friday'),
      6 => t('Saturday'),
      0 => t('Sunday'),
    ];

    $weekday_start = $this->options['calendar_weekday_start'] ?? 0;
    $weekdays = range($weekday_start, 6);
    $days = array_replace(array_flip($weekdays), $days);

    $headers = [];
    foreach ($days as $number => $name) {
      $headers[$number] = $name;
    }

    // Dates for this month.
    $month_start = strtotime($year . '-' . $month . '-' . '01');
    $month_days = date('t', $month_start);
    $first_day = date('w', $month_start);
    $month_weeks = floor(($first_day + $month_days - 1) / 7);
    $month_end = strtotime($year . '-' . $month . '-' . $month_days);

    // Next month.
    $next_month = $month == '12' ? '01' : str_pad(($month + 1), 2, '0', STR_PAD_LEFT);
    $next_year = $month == '12' ? $year + 1 : $year;

    // Last month.
    $previous_month = $month == '01' ? '12' : str_pad(($month - 1), 2, '0', STR_PAD_LEFT);
    $previous_year = $month == '01' ? $year - 1 : $year;
    $previous_month_start = strtotime($previous_year . '-' . $previous_month . '-' . '01');
    $previous_month_days = date('t', $previous_month_start);

    $previous_month_offset = [];
    $month_started = FALSE;
    foreach (array_keys($headers) as $number) {
      if ($month_started = (int) $number == (int) $first_day) {
        break;
      }
      $previous_month_offset[] = $previous_month_days;
      $previous_month_days--;
    }
    $previous_month_offset = array_reverse($previous_month_offset);

    $count = 0;
    $month_started = FALSE;
    for ($i = 0; $i <= $month_weeks; $i++) {
      // Prepare row.
      $cells = [];

      // First week.
      if ($i == 0) {
        // Empty days starting the month display.
        foreach ($previous_month_offset as $daynum) {
          $day_number = str_pad($daynum, 2, '0', STR_PAD_LEFT);
          $time_now = strtotime($previous_year . '-' . $previous_month . '-' . $day_number);

          $cells[$time_now]['data'] = [
            '#theme' => 'calendar_view_day',
            '#timestamp' => $time_now,
            '#children' => [],
          ];
          $cells[$time_now]['data-calendar-view-day'] = date('d', $time_now);
          $cells[$time_now]['data-calendar-view-month'] = date('m', $time_now);
          $cells[$time_now]['data-calendar-view-year'] = date('y', $time_now);
          $cells[$time_now]['class'] = ['previous-month'];
        }

        // Pending days of this month's first week.
        $x = 7 - count($previous_month_offset);
        do {
          $x--;

          // Count days of the month.
          $count++;

          // Get this day's timestamp.
          $day_number = str_pad($count, 2, '0', STR_PAD_LEFT);
          $time_now = strtotime($year . '-' . $month . '-' . $day_number);

          $cells[$time_now]['data'] = [
            '#theme' => 'calendar_view_day',
            '#timestamp' => $time_now,
            '#children' => [],
          ];
          $cells[$time_now]['data-calendar-view-day'] = date('d', $time_now);
          $cells[$time_now]['data-calendar-view-month'] = date('m', $time_now);
          $cells[$time_now]['data-calendar-view-year'] = date('y', $time_now);
          $cells[$time_now]['class'] = ['current-month'];
        } while ($x >= 1);

        // Populate table row.
        $rows[] = ['data' => $cells];

        continue;
      }

      // Rest of the weeks.
      $daynum = 0;
      foreach (array_keys($headers) as $number) {
        // Count days of the month.
        $count++;

        // Fill next months day, if necessary.
        $month_finished = $count > (int) $month_days;
        $week_finished = $number == count($headers);
        if ($month_finished && !$week_finished) {
          $daynum++;
          $time_now = strtotime($next_year . '-' . $next_month . '-' . $daynum);
          $cells[$time_now]['data'] = [
            '#theme' => 'calendar_view_day',
            '#timestamp' => $time_now,
            '#children' => [],
          ];
          $cells[$time_now]['data-calendar-view-day'] = date('d', $time_now);
          $cells[$time_now]['data-calendar-view-month'] = date('m', $time_now);
          $cells[$time_now]['data-calendar-view-year'] = date('y', $time_now);
          $cells[$time_now]['class'] = ['next-month'];
          continue;
        }

        // Stop now.
        if ($month_finished) {
          break;
        }

        // Insert day.
        $day_number = str_pad($count, 2, '0', STR_PAD_LEFT);
        $time_now = strtotime($year . '-' . $month . '-' . $day_number);

        $cells[$time_now]['data'] = [
          '#theme' => 'calendar_view_day',
          '#timestamp' => $time_now,
          '#children' => [],
        ];
        $cells[$time_now]['data-calendar-view-day'] = date('d', $time_now);
        $cells[$time_now]['data-calendar-view-month'] = date('m', $time_now);
        $cells[$time_now]['data-calendar-view-year'] = date('y', $time_now);
        $cells[$time_now]['class'] = ['current-month'];
      }

      // Populate table row.
      $rows[] = ['data' => $cells];
    }

    return [
      '#type' => 'table',
      '#caption' => date('F Y', $month_start),
      '#header' => $headers,
      '#rows' => $rows,
      '#empty' => NULL,
      '#attributes' => [
        'data-calendar-view-year' => $year,
        'data-calendar-view-month' => $month,
        'class' => [
          'calendar-view-table',
        ],
      ],
    ];
  }
}
